/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2020 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system for microcontrollers      *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.10.2.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
File    : BSP_UART.c
Purpose : UART implementation

Additional information:

  Device : MK66F18
  Board  : SEGGER emPower

  Unit | UART   | Board connector
  ===============================
  0    | UART0  | Serial A (J25)
  1    | UART1  | RS-232   (J6)
  2    | UART2  | -

*/

#include "BSP_UART.h"
#include "RTOS.h"      // For OS_INT_Enter()/OS_INT_Leave(). Remove this line and OS_INT_* functions if not using OS.

/*********************************************************************
*
*       Defines, fixed
*
**********************************************************************
*/
#define UMR1_0       (*(volatile unsigned char*)0x40000200)
#define UMR2_0       (*(volatile unsigned char*)0x40000200)
#define USR_0        (*(volatile unsigned char*)0x40000204)
#define UCSR_0       (*(volatile unsigned char*)0x40000240)
#define UCR_0        (*(volatile unsigned char*)0x40000208)
#define URB_0        (*(volatile unsigned char*)0x4000020C)
#define UTB_0        (*(volatile unsigned char*)0x4000020C)
#define UIPCR_0      (*(volatile unsigned char*)0x40000210)
#define UACR_0       (*(volatile unsigned char*)0x40000210)
#define UISR_0       (*(volatile unsigned char*)0x40000214)
#define UIMR_0       (*(volatile unsigned char*)0x40000214)
#define UIP_0        (*(volatile unsigned char*)0x40000234)
#define UBG0_0       (*(volatile unsigned char*)0x40000218)
#define UBG1_0       (*(volatile unsigned char*)0x4000021C)
#define UOP0_0       (*(volatile unsigned char*)0x40000238)
#define UOP1_0       (*(volatile unsigned char*)0x40000238)
#define UGPIO        (*(volatile unsigned char*)0x40100071)
#define OS_BAUDRATE  (38400u)
#define BAUDRATEREG  ((80000000u / 1000u * 125u) / (OS_BAUDRATE * 4u))
#define OS_UART_ID   (13u)

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/
static BSP_UART_TX_CB* _pfWriteCB;
static BSP_UART_RX_CB* _pfReadCB;
static unsigned int    _TxIsBusy;  // Used for a combined Rx & Tx interrupt that uses a TxEmpty logic.

/*********************************************************************
*
*       Prototypes
*
*  Declare ISR handler here to avoid "no prototype" warning.
*  They are not declared in any CMSIS header.
*
**********************************************************************
*/

#if defined(__cplusplus)
  extern "C" {                // Make sure we have C-declarations in C++ programs.
#endif

void UART_RX_TX_IRQHandler(void);

#if defined(__cplusplus)
}                             // Make sure we have C-declarations in C++ programs.
#endif

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _SetBaudrate()
*
*  Function description
*    Configures the UART baud rate.
*
*  Parameters
*    Unit    : Unit number (typically zero-based).
*    Baudrate: Baud rate to configure [Hz].
*/
static void _SetBaudrate(unsigned int Unit, unsigned long Baudrate) {
  BSP_UART_USE_PARA(Unit);
  BSP_UART_USE_PARA(Baudrate);
}

/*********************************************************************
*
*       Global functions, IRQ handler
*
**********************************************************************
*/

/*********************************************************************
*
*       UART_RX_TX_IRQHandler()
*
*  Function description
*    UART Rx & Tx interrupt handler.
*
*  Additional information
*    Needs to inform embOS that we are in interrupt context.
*/
void UART_RX_TX_IRQHandler(void) {
  unsigned char Data;
  unsigned int  Unit;

  Unit = 0u;                        // Identify UART unit if multiple units are used.

  if (USR_0 & 0x01u) {              // Rx interrupt
    Data = URB_0;
    if ((USR_0 & 0xF0u) == 0x00) {  // Check if errors occurred
      if (_pfReadCB != NULL) {
        _pfReadCB(Unit, Data);      // Forward the received byte to the callback.
      }
    } else {
      UCR_0 |= (4u << 4);           // Reset Rx errors
      UCR_0 &= 0x0Fu;
    }
  }
  if (UISR_0 & 0x01u) {             // Tx interrupt
    if (_pfWriteCB != NULL) {
      if (_pfWriteCB(Unit) != 0) {  // Tell the callback that we are ready to send the next byte.
        UIMR_0 = 0x02u;             // Disable Tx interrupt (Rx interrupt is still enabled)
      }
    }
  }
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       BSP_UART_Init()
*
*  Function description
*    Initializes the UART peripheral.
*
*  Parameters
*    Unit       : Unit number (typically zero-based).
*    Baudrate   : Baud rate to configure [Hz].
*    NumDataBits: Number of data bits to use.
*    Parity     : One of the following values:
*                   * BSP_UART_PARITY_NONE
*                   * BSP_UART_PARITY_ODD
*                   * BSP_UART_PARITY_EVEN
*    NumStopBits: Number of stop bits to use.
*/
void BSP_UART_Init(unsigned int Unit, unsigned long Baudrate, unsigned char NumDataBits, unsigned char Parity, unsigned char NumStopBits) {
  _TxIsBusy = 0u;

  UGPIO  = 0x05;                       // Init GPIO pins.
  UCR_0  = 0x30;                       // Reset transmitter
  UCR_0  = 0x20;                       // Reset reciver
  UCR_0  = 0x10;                       // Reset Mode Register
  UMR1_0 = 0x13;                       // 8N1
  UMR2_0 = 0x07;
  UCSR_0 = 0xDD;                       // Set Rx and Tx baud by SYSTEM CLOCK
  UIMR_0 = 0x02;                       // Enable Rx interrupts
  UBG0_0 = (OS_U8)(BAUDRATEREG >> 8);  // Set baud settings
  UBG1_0 = (OS_U8)BAUDRATEREG;
  UCR_0  = 0x05;                       // Enable receiver and transmitter
  //
  // Configure NumDataBits, Parity and NumStopBits.
  //
  BSP_UART_USE_PARA(Unit);
  BSP_UART_USE_PARA(Baudrate);
  BSP_UART_USE_PARA(NumDataBits);
  BSP_UART_USE_PARA(Parity);
  BSP_UART_USE_PARA(NumStopBits);
  //
  // Setup UART interrupt(s).
  //
  OS_EnableISR(OS_UART_ID, UART_RX_TX_IRQHandler, 2, 0);
}

/*********************************************************************
*
*       BSP_UART_DeInit()
*
*  Function description
*    De-initializes the UART peripheral.
*
*  Parameters
*    Unit: Unit number (typically zero-based).
*/
void BSP_UART_DeInit(unsigned int Unit) {
  BSP_UART_USE_PARA(Unit);

  OS_DisableISR(OS_UART_ID);
}

/*********************************************************************
*
*       BSP_UART_SetBaudrate()
*
*  Function description
*    Configures/changes the UART baud rate.
*
*  Parameters
*    Unit    : Unit number (typically zero-based).
*    Baudrate: Baud rate to configure [Hz].
*/
void BSP_UART_SetBaudrate(unsigned int Unit, unsigned long Baudrate) {
  _SetBaudrate(Unit, Baudrate);
}

/*********************************************************************
*
*       BSP_UART_SetReadCallback()
*
*  Function description
*    Sets the callback to execute upon an Rx interrupt.
*
*  Parameters
*    Unit: Unit number (typically zero-based).
*    pf  : Callback to execute.
*/
void BSP_UART_SetReadCallback(unsigned Unit, BSP_UART_RX_CB* pf) {
  BSP_UART_USE_PARA(Unit);

  _pfReadCB = pf;
}

/*********************************************************************
*
*       BSP_UART_SetWriteCallback()
*
*  Function description
*    Sets the callback to execute upon a Tx interrupt.
*
*  Parameters
*    Unit: Unit number (typically zero-based).
*    pf  : Callback to execute.
*/
void BSP_UART_SetWriteCallback(unsigned int Unit, BSP_UART_TX_CB* pf) {
  BSP_UART_USE_PARA(Unit);

  _pfWriteCB = pf;
}

/*********************************************************************
*
*       BSP_UART_Write1()
*
*  Function description
*    Sends one byte via UART.
*
*  Parameters
*    Unit: Unit number (typically zero-based).
*    Data: (First) data byte to send.
*
*  Additional information
*    The first byte of a transfer is typically sent from application
*    context. Further bytes of the transfer are then sent from the
*    Tx interrupt handler by also calling this function from interrupt
*    context.
*/
void BSP_UART_Write1(unsigned int Unit, unsigned char Data) {
  BSP_UART_USE_PARA(Unit);
  BSP_UART_USE_PARA(Data);

  UTB_0  = Data;  // store character in tx register
  UIMR_0 = 0x03;  // enable Tx and Rx interrupt
}

/*************************** End of file ****************************/
