/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2020 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system for microcontrollers      *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.10.2.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
File    : RTOSInit.c
Purpose : Initializes and handles the hardware for embOS
*/

#include "RTOS.h"

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/

/*********************************************************************
*
*       System tick settings
*/
#define OS_TIMER_FREQ  (40000000u)
#define OS_TICK_FREQ   (1000u)
#define OS_INT_FREQ    (OS_TICK_FREQ)

#ifndef   OS_ISRHANDLERTABLE_RAM
  #define OS_ISRHANDLERTABLE_RAM  (0)
#endif

/*********************************************************************
*
*       embOSView settings
*/
#ifndef   OS_VIEW_IFSELECT
  #define OS_VIEW_IFSELECT  OS_VIEW_IF_UART
#endif

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  #include "BSP_UART.h"
  #define OS_UART      (0u)
  #define OS_BAUDRATE  (38400u)
#endif

#define NUM_INT_SOURCES           (64u)
#define OS_ID_SYSTICK             (55u)
//
// SFRs definition for PIT0
//
#define PIT0_PCSR                 (*(volatile unsigned short*) 0x40150000)
#define PIT0_PMR                  (*(volatile unsigned short*) 0x40150002)
#define PIT0_CNTR                 (*(volatile unsigned short*) 0x40150004)
#define PIT_PCSR_PIF              (0x0004u)
//
// SFRs definition for PLL
//
#define MCF_CLOCK_SYNCR           (*(volatile unsigned short*) 0x40120000)
#define MCF_CLOCK_SYNSR           (*(volatile unsigned char*)  0x40120002)
#define MCF_CLOCK_CCHR            (*(volatile unsigned char*)  0x40120008)
//
// Bit definitions and macros for MCF_CLOCK_SYNCR
//
#define MCF_CLOCK_SYNCR_PLLEN     (0x0001u)
#define MCF_CLOCK_SYNCR_PLLMODE   (0x0002u)
#define MCF_CLOCK_SYNCR_CLKSRC    (0x0004u)
#define MCF_CLOCK_SYNCR_LOCRE     (0x0800u)
#define MCF_CLOCK_SYNCR_MFD(x)    (((x) & 0x0007u ) << 12u)
//
// Bit definitions and macros for MCF_CLOCK_SYNSR
//
#define MCF_CLOCK_SYNSR_LOCK      (0x08u)
//
// SFR definitions for interrupt controller
//
#define ICR_BASE_ADDR             (0x40000C40u)
#define IMRH                      (*(volatile unsigned long*)0x40000C08u)
#define IMRL                      (*(volatile unsigned long*)0x40000C0Cu)

/*********************************************************************
*
*       Prototypes
*
**********************************************************************
*/
static void _OS_IrqDefaultHandler(void);
static void SysTick_Handler      (void);

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/
#if (OS_ISRHANDLERTABLE_RAM == 1)
static OS_ISR_HANDLER* _papISR[NUM_INT_SOURCES];
#else
static OS_ISR_HANDLER* const _papISR[NUM_INT_SOURCES] = {
  _OS_IrqDefaultHandler,      // Handler for interrupt 0
  _OS_IrqDefaultHandler,      // Handler for interrupt 1
  _OS_IrqDefaultHandler,      // Handler for interrupt 2
  _OS_IrqDefaultHandler,      // Handler for interrupt 3
  _OS_IrqDefaultHandler,      // Handler for interrupt 4
  _OS_IrqDefaultHandler,      // Handler for interrupt 5
  _OS_IrqDefaultHandler,      // Handler for interrupt 6
  _OS_IrqDefaultHandler,      // Handler for interrupt 7
  _OS_IrqDefaultHandler,      // Handler for interrupt 8
  _OS_IrqDefaultHandler,      // Handler for interrupt 9
  _OS_IrqDefaultHandler,      // Handler for interrupt 10
  _OS_IrqDefaultHandler,      // Handler for interrupt 11
  _OS_IrqDefaultHandler,      // Handler for interrupt 12
#if (OS_UART_USED)
  _Uart0Handler,              // Handler for interrupt 13, communication to embOSView
#else
  _OS_IrqDefaultHandler,      // Handler for interrupt 13
#endif
  _OS_IrqDefaultHandler,      // Handler for interrupt 14
  _OS_IrqDefaultHandler,      // Handler for interrupt 15
  _OS_IrqDefaultHandler,      // Handler for interrupt 16
  _OS_IrqDefaultHandler,      // Handler for interrupt 17
  _OS_IrqDefaultHandler,      // Handler for interrupt 18
  _OS_IrqDefaultHandler,      // Handler for interrupt 19
  _OS_IrqDefaultHandler,      // Handler for interrupt 20
  _OS_IrqDefaultHandler,      // Handler for interrupt 21
  _OS_IrqDefaultHandler,      // Handler for interrupt 22
  _OS_IrqDefaultHandler,      // Handler for interrupt 23
  _OS_IrqDefaultHandler,      // Handler for interrupt 24
  _OS_IrqDefaultHandler,      // Handler for interrupt 25
  _OS_IrqDefaultHandler,      // Handler for interrupt 26
  _OS_IrqDefaultHandler,      // Handler for interrupt 27
  _OS_IrqDefaultHandler,      // Handler for interrupt 28
  _OS_IrqDefaultHandler,      // Handler for interrupt 29
  _OS_IrqDefaultHandler,      // Handler for interrupt 30
  _OS_IrqDefaultHandler,      // Handler for interrupt 31
  _OS_IrqDefaultHandler,      // Handler for interrupt 32
  _OS_IrqDefaultHandler,      // Handler for interrupt 33
  _OS_IrqDefaultHandler,      // Handler for interrupt 34
  _OS_IrqDefaultHandler,      // Handler for interrupt 35
  _OS_IrqDefaultHandler,      // Handler for interrupt 36
  _OS_IrqDefaultHandler,      // Handler for interrupt 37
  _OS_IrqDefaultHandler,      // Handler for interrupt 38
  _OS_IrqDefaultHandler,      // Handler for interrupt 39
  _OS_IrqDefaultHandler,      // Handler for interrupt 40
  _OS_IrqDefaultHandler,      // Handler for interrupt 41
  _OS_IrqDefaultHandler,      // Handler for interrupt 42
  _OS_IrqDefaultHandler,      // Handler for interrupt 43
  _OS_IrqDefaultHandler,      // Handler for interrupt 44
  _OS_IrqDefaultHandler,      // Handler for interrupt 45
  _OS_IrqDefaultHandler,      // Handler for interrupt 46
  _OS_IrqDefaultHandler,      // Handler for interrupt 47
  _OS_IrqDefaultHandler,      // Handler for interrupt 48
  _OS_IrqDefaultHandler,      // Handler for interrupt 49
  _OS_IrqDefaultHandler,      // Handler for interrupt 50
  _OS_IrqDefaultHandler,      // Handler for interrupt 51
  _OS_IrqDefaultHandler,      // Handler for interrupt 52
  _OS_IrqDefaultHandler,      // Handler for interrupt 53
  _OS_IrqDefaultHandler,      // Handler for interrupt 54
  SysTick_Handler,            // Handler for interrupt 55, the OS system tick
  _OS_IrqDefaultHandler,      // Handler for interrupt 56
  _OS_IrqDefaultHandler,      // Handler for interrupt 57
  _OS_IrqDefaultHandler,      // Handler for interrupt 58
  _OS_IrqDefaultHandler,      // Handler for interrupt 59
  _OS_IrqDefaultHandler,      // Handler for interrupt 60
  _OS_IrqDefaultHandler,      // Handler for interrupt 61
  _OS_IrqDefaultHandler,      // Handler for interrupt 62
  _OS_IrqDefaultHandler,      // Handler for interrupt 63
};
#endif

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_IrqDefaultHandler()
*
*  Function description
*    Default IRQ handler
*/
static void _OS_IrqDefaultHandler(void) {
#if (OS_DEBUG != 0)
  OS_INT_Disable();
  while (1) {
    //
    // The NOPs are necessary for bdm debugger,
    // otherwise the bdm debugger locks up
    //
    asm(NOP);
    asm(NOP);
  }
#endif
}

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
/*********************************************************************
*
*       _OS_OnRX()
*
*  Function description
*    Callback wrapper function for BSP UART module.
*/
static void _OS_OnRX(unsigned int Unit, unsigned char c) {
  OS_USEPARA(Unit);
  OS_COM_OnRx(c);
}

/*********************************************************************
*
*       _OS_OnTX()
*
*  Function description
*    Callback wrapper function for BSP UART module.
*/
static int _OS_OnTX(unsigned int Unit) {
  OS_USEPARA(Unit);
  return (int)OS_COM_OnTx();
}
#endif

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
*  Function description
*    Returns the current hardware timer count value.
*
*  Return value
*    Current timer count value.
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  return PIT0_CNTR;
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
*  Function description
*    Returns if the hardware timer interrupt pending flag is set.
*
*  Return value
*    == 0: Interrupt pending flag not set.
*    != 0: Interrupt pending flag set.
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  return (PIT0_PCSR & PIT_PCSR_PIF);
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       SysTick_Handler()
*
*  Function description
*    This is the hardware timer exception handler.
*/
void SysTick_Handler(void) {
  PIT0_PCSR |= PIT_PCSR_PIF;  // Reset timer interrupt flag
  OS_TICK_Handle();
}

/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*    Initialize the hardware required for embOS to run.
*/
const OS_SYSTIMER_CONFIG SysTimerConfig = {OS_TIMER_FREQ, OS_INT_FREQ, OS_TIMER_DOWNCOUNTING, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};
void OS_InitHW(void) {
#if (OS_ISRHANDLERTABLE_RAM == 1)
  unsigned int i;
#endif

  OS_INT_IncDI();
  //
  // Inform embOS about the timer settings
  //
  OS_TIME_ConfigSysTimer(&SysTimerConfig);
  //
  // Initialize the vector table
  //
#if (OS_ISRHANDLERTABLE_RAM == 1)
  for (i = 0; i < NUM_INT_SOURCES; +i++) {
    _papISR[i] = _OS_IrqDefaultHandler;
  }
#endif
  //
  // Initialize PLL
  //
  MCF_CLOCK_CCHR  = 0x05;
  MCF_CLOCK_SYNCR = MCF_CLOCK_SYNCR_MFD(3) | MCF_CLOCK_SYNCR_CLKSRC| MCF_CLOCK_SYNCR_PLLMODE | MCF_CLOCK_SYNCR_PLLEN;
  while ((MCF_CLOCK_SYNSR & MCF_CLOCK_SYNSR_LOCK) == 0) {
  }
  //
  // Init PIT0 to generate periodic interrupt = 1ms
  //
  OS_EnableISR(OS_ID_SYSTICK, SysTick_Handler, 1, 0);
  PIT0_PCSR  = 0x0000u;                                         // Prescaler 2^0=1
  PIT0_PMR   = (unsigned short)(OS_TIMER_FREQ / OS_TICK_FREQ);  // Reload value for timer
  PIT0_PCSR |=  (1u << 0)                                       // 1 = enabled EN
              | (1u << 1)                                       // 1 = counter reload RLD
              | (0u << 2)                                       // Int reset bit PIF
              | (1u << 3)                                       // Overwrite bit PIE
              | (1u << 4)                                       // Overwrite bit OVW
              | (0u << 5)                                       // Debug mode
              | (0u << 6)                                       // Doze mode bit
              | (0u << 7);                                      // Reserved
  //
  // Initialize communication for embOSView
  //
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Init(OS_UART, OS_BAUDRATE, BSP_UART_DATA_BITS_8, BSP_UART_PARITY_NONE, BSP_UART_STOP_BITS_1);
  BSP_UART_SetReadCallback(OS_UART, _OS_OnRX);
  BSP_UART_SetWriteCallback(OS_UART, _OS_OnTX);
#endif
  OS_INT_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*  Function description
*    This code is executed whenever no task, software timer, or
*    interrupt is ready for execution.
*
*  Additional information
*    The idle loop does not have a stack of its own, therefore no
*    functionality should be implemented that relies on the stack
*    to be preserved.
*/
void OS_Idle(void) {  // Idle loop: No task is ready to execute
  while (1) {         // Nothing to do ... wait for interrupt
    //
    // The NOPs are necessary for the bdm debugger,
    // otherwise the bdm debugger locks up.
    //
    asm(NOP);
    asm(NOP);
  }
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character.
*/
void OS_COM_Send1(OS_U8 c) {
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Write1(OS_UART, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_DISABLED)
  OS_USEPARA(c);           // Avoid compiler warning
  OS_COM_ClearTxActive();  // Let embOS know that Tx is not busy
#endif
}

/*********************************************************************
*
*       OS interrupt handler and ISR specific functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_irq_handler()
*
*  Function description
*    Detect reason for IRQ and call corresponding service routine.
*    OS_irq_handler is called from OS_ISR_Handler function
*    from the embOS library
*/
void OS_irq_handler(unsigned int ISRIndex) {
  OS_ISR_HANDLER* pISR;

#if (OS_DEBUG != 0)
  if (ISRIndex < 64u) {
    //
    // OS_irq_Handler() handles peripheral interrupts only
    //
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  pISR = _papISR[ISRIndex - 64u];
  pISR();
}

/*********************************************************************
*
*       OS_EnableISR()
*
*  Function description
*    Installs and enables an ISR.
*/
OS_ISR_HANDLER* OS_EnableISR(unsigned int ISRIndex, OS_ISR_HANDLER* pISRHandler, unsigned int Priority, unsigned int SubPriority) {
  OS_ISR_HANDLER*  pOldHandler;
#if (OS_DEBUG != 0)
  unsigned int i;

  if ((ISRIndex >= NUM_INT_SOURCES) || (ISRIndex == 0)) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  pOldHandler = _papISR[ISRIndex];             // Get old handler

#if (OS_ISRHANDLERTABLE_RAM == 1)
  _papISR[ISRIndex] = pISRHandler;             // Install new handler
#else
  OS_USEPARA(pISRHandler);                     // Not used, avoid warning
#endif
  //
  // Check if the interrupt level/priority is unique.
  // Failure to program the ICRnx registers in this manner can result in undefined behavior.
  //
#if (OS_DEBUG != 0)
  for (i = 8; i < NUM_INT_SOURCES; i++) {
    if (i != ISRIndex) {
      if (*(char*) (ICR_BASE_ADDR + i) == ((Priority << 3) | SubPriority)) {
        OS_Error(OS_ERR_ISR_VECTOR);
      }
    }
  }
#endif
  *(char*)(ICR_BASE_ADDR + ISRIndex) = (char)((Priority << 3) | SubPriority);   // Set IPL register
  if ( ISRIndex < 32) {                        // Set interrupt mask register IMRL, IMRH
    IMRL &= ~((1uL << ISRIndex) | 0x0001);
  } else {
    IMRL &= ~(0x0001uL);                       // IMRL.bit0 has to be reset to enable any interrupts
    IMRH &= ~(1uL << (ISRIndex-32));
  }
  OS_INT_EnableConditional();
  return pOldHandler;
}

/*********************************************************************
*
*       OS_DisableISR()
*
*  Function description
*    Disables an ISR.
*/
void OS_DisableISR(unsigned int ISRIndex) {
#if (OS_DEBUG != 0)
  if (ISRIndex >= NUM_INT_SOURCES || (ISRIndex == 0)) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  if ( ISRIndex < 32) {                        // Reset interrupt mask register
    IMRL |= (1uL << ISRIndex);
  } else {
    IMRH |= (1uL << (ISRIndex - 32));
  }
  *(char*)(ICR_BASE_ADDR + ISRIndex) = 0x00;   // Set IPL register to 0
#if (OS_ISRHANDLERTABLE_RAM == 1)
  _papISR[ISRIndex] = _OS_IrqDefaultHandler;   // Set default handler
#endif
  OS_INT_EnableConditional();
}

/*************************** End of file ****************************/
